/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/


/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include "../EM/atmega128.h"
#include "opcodes.h"
#include "../EM/debug.h"

error_t atmega_op_ldi( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "LDI\tR%d, 0x%02x\t;%d\n", chip->reg_index_a, chip->b, chip->b );
        return 1;
    }
	
    chip->register_file[chip->reg_index_a] = chip->b;
	
    chip->cycles += 1;
    chip->pc++;
	
    return 0;
}

error_t atmega_op_in( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "IN\tR%d, 0x%02x\n", chip->reg_index_a, chip->b);
        return 1;
    }
	
    chip->pc+=1;
    chip->cycles+=1;
	
    return atmega_read_sram(chip, chip->b + chip->io_offset, chip->register_file+chip->reg_index_a);
}

error_t atmega_op_out( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "OUT\t0x%02x, R%d\n", chip->b, chip->reg_index_a );
        return 1;
    }
	
    atmega_write_sram( chip, chip->b + chip->io_offset, chip->a );
	
    chip->cycles += 1;
    chip->pc++;
	
    return 0;
}
char g_tempLdDbg=0;
/* whj rewrite this function!
error_t atmega_op_ld_ldd( atmega_chip *chip )
{
int inc;  // amount to increment by /
uint16_t off; // amount to offset by /
uint8_t reg;
uint8_t preg_l, preg_h; // pointer register /
uint16_t addr;
char preg_name;

  reg=chip->reg_index_a;
  
    if (((chip->bl_nibble >> 2) == 3) && (chip->th_nibble == 0x9))
    {
	preg_l=ATEMU_REG_X_LOW;
	preg_h=ATEMU_REG_X_HIGH;
	preg_name='X';
    }else if((chip->bl_nibble >> 3) == 1){
	preg_l=ATEMU_REG_Y_LOW;
	preg_h=ATEMU_REG_Y_HIGH;
	preg_name='Y';
    }else if((chip->bl_nibble >> 3) == 0){
	preg_l=ATEMU_REG_Z_LOW;
	preg_h=ATEMU_REG_Z_HIGH;
	preg_name='Z';
    }else{
	// unknown pointer register /
	log_error(chip, "error: unknown pointer register\n");
	return -1;
    }
	
	  addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
	  
		if(chip->th_nibble==8 || chip->th_nibble==10){
        inc=0;
        off=chip->b;
        if(chip->mode == ATEMU_MODE_DECODE){
		if(off != 0){
		log_opcode(chip, "LDD\tR%d, %c%+d\n", reg, preg_name, off);
		}else{
		log_opcode(chip, "LD\tR%d, %c\n", reg, preg_name);
		}
		return 1;
        }
		}else if((chip->bl_nibble & 3) == 0){
        inc=0;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
		log_opcode(chip, "LD\tR%d, %c\n", reg, preg_name);
		return 1;
        }
		}else if((chip->bl_nibble & 3) == 1){
        inc=1;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
		log_opcode(chip, "LD\tR%d, %c+\n", reg, preg_name);
		return 1;
        }
		}else if((chip->bl_nibble & 3) == 2){
        inc=-1;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
		log_opcode(chip, "LD\tR%d, -%c\n", reg, preg_name);
		return 1;
        }
		}else{
        log_error(chip, "Unknown increment setting\n");
        return -1;
		}
		
		  if(inc == -1){
		  addr--;
		  }
		  atmega_read_sram(chip, addr + off, &chip->register_file[reg]);
		  if(inc == 1){
		  addr++;
		  }
		  if((addr + off)==0x3a7||(addr + off)==0x3a6)
		  {
		  g_tempLdDbg=1-g_tempLdDbg;
		  
			}
			chip->register_file[preg_h]=HIGH_BYTE_16(addr);
			chip->register_file[preg_l]=LOW_BYTE_16(addr);
			
			  
				//if(logfile && logmemory){
				// fprintf(logfile, "%03d LDD %15ld\t%x\t%d\t%x\t%x\n", chip->id, (long)chip->cycles,
				// (int)chip->pc, (int)reg, (unsigned int)addr, (unsigned int)chip->register_file[reg]);
				//}/
				//
				
				  chip->cycles += 2;
				  chip->pc++;
				  
					return 0;
					}
*/

//////////////////////////////////////////////////////////////////////////////////////
//VMNet: atmega_op_ld_ldd
//This function is rewrited as the previous function in atemu is problemsitc.
//The previous function (above) always increase the register pointer.
// However, it is wrong when the code in th_nibble is0x8 or oxa, the address register need 
// not change. If always increase or decrease thevalue in registerXYZ may be wrong.

///////////////////////////////////////////////////////////////////////////////////////////////
static unsigned char ld_th_mask_table[]={0x9,0x8,0xa};
static unsigned char ld_tl_mask_table[]={0xd,0x8,0x2};
static unsigned char ld_bl_mask_table[]={0xc,0x8,0x3};

error_t atmega_op_ld_ldd( atmega_chip *chip )
{
    int inc=0;  /* amount to increment by */
    uint16_t off=0; /* amount to offset by */
    uint8_t reg;
    uint8_t preg_l=0xf, preg_h=0x1; /* pointer register */
    uint16_t addr;
    char preg_name;
	
    reg=chip->reg_index_a;
    switch(chip->th_nibble)
    {
	case 0x9: // class 9 increament/decreament
		if(chip->bl_nibble>=ld_bl_mask_table[0])//X
		{
			preg_l=ATEMU_REG_X_LOW;
			preg_h=ATEMU_REG_X_HIGH;
			addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
			preg_name='X';
			if(chip->bl_nibble==0xc)
			{
				// operation
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
			}
			else if(chip->bl_nibble==0xd)
			{
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
				addr++;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
			}
			else if(chip->bl_nibble==0xe)
			{
				addr--;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
				atmega_read_sram(chip, addr, &chip->register_file[reg]);                  
			}
			else
			{
				log_error(chip, "error: unknown pointer register\n");
				return -1;
			}
		}  //X over
		else if(chip->bl_nibble>=ld_bl_mask_table[1])//Y
		{
			preg_l=ATEMU_REG_Y_LOW;
			preg_h=ATEMU_REG_Y_HIGH;
			addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
			preg_name='Y';
			if(chip->bl_nibble==0x9)
			{
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
				addr++;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
			}
			else if(chip->bl_nibble==0xa)
			{
				addr--;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
			}
			else
			{
				log_error(chip, "error: unknown pointer register\n");
				return -1;
			}
			
		}// 2 instructs of Y
		else //Z
		{
			preg_l=ATEMU_REG_Z_LOW;
			preg_h=ATEMU_REG_Z_HIGH;
			addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
			preg_name='Z';
			if(chip->bl_nibble==0x1)
			{
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
				addr++;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
			}
			else if(chip->bl_nibble==0x2)
			{
				addr--;
				chip->register_file[preg_h]=HIGH_BYTE_16(addr);
				chip->register_file[preg_l]=LOW_BYTE_16(addr);
				atmega_read_sram(chip, addr, &chip->register_file[reg]);
			}
			else
			{
				log_error(chip, "error: unknown pointer register\n");
				return -1;
			} //2 instructs of Z
			
		}
		break;
	case 0x8:// class 8 unchange
	case 0xa:
		off=chip->b;
		if(chip->bl_nibble>=ld_bl_mask_table[1])//Y
		{
			preg_l=ATEMU_REG_Y_LOW;
			preg_h=ATEMU_REG_Y_HIGH;
			addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
			preg_name='Y';
			atmega_read_sram(chip, addr+off, &chip->register_file[reg]);      
		}
		else
		{
			preg_l=ATEMU_REG_Z_LOW;
			preg_h=ATEMU_REG_Z_HIGH;
			addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
			preg_name='Z';
			atmega_read_sram(chip, addr+off, &chip->register_file[reg]);      
			
		}
		break;
	default:
        log_error(chip, "error: unknown pointer register\n");
        return -1;      
    }//switch
    chip->cycles += 2;
    chip->pc++;
	
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////

error_t atmega_op_lpm(atmega_chip *chip){
    uint16_t addr;     /* source address for read */
    uint32_t regz_l, regz_h; /* parts of address */
    uint16_t data;
    uint8_t dreg;
    int increment;
	
    switch(chip->bl_nibble){
	case 0x4:
		dreg=chip->reg_index_a;
		increment=0;
		break;
	case 0x5:
		dreg=chip->reg_index_a;
		increment=1;
		break;
	case 0x8:
		dreg=0;
		increment=0;
		break;
	default:
		log_error(chip, "Unknown destination register and increment setting!\n");
		return -1;
    }
	
    if(chip->mode == ATEMU_MODE_DECODE){
        if(increment){
            log_opcode(chip,  "LPM\tR%d, Z+\n", dreg );
        }else{
            log_opcode(chip,  "LPM\tR%d, Z\n", dreg );
        }
        return 1;
    }
	
    regz_l=chip->register_file[ATEMU_REG_Z_LOW];
    regz_h=chip->register_file[ATEMU_REG_Z_HIGH];
    addr=PACK_INT_16(regz_h, regz_l);
	
    /* addr is a byte address, flash is word addressable */
    data=(chip->flash[addr >> 1]);
    data=(chip->flash[addr >> 1]);
    if(addr & 1){
        chip->register_file[dreg]=TOP_BYTE(data);
    }else{
        chip->register_file[dreg]=BOTTOM_BYTE(data);
    }
	
    if(increment){
        /* increment addr, unpack, write back */
        addr++;
        regz_l=LOW_BYTE_16(addr);
        regz_h=HIGH_BYTE_16(addr);
        chip->register_file[ATEMU_REG_Z_LOW]=regz_l;
        chip->register_file[ATEMU_REG_Z_HIGH]=regz_h;
    }
	
    chip->cycles += 3;
    chip->pc++;
	
    return 0;
}
error_t atmega_op_elpm(atmega_chip *chip){
    uint32_t addr;     /* source address for read */
    uint32_t rampz, regz_l, regz_h; /* parts of address */
    uint16_t data;
    uint8_t dreg;
    int increment;
	
    switch(chip->bl_nibble){
	case 0x6:
		dreg=chip->reg_index_a;
		increment=0;
		break;
	case 0x7:
		dreg=chip->reg_index_a;
		increment=1;
		break;
	case 0x8:
		dreg=0;
		increment=0;
		break;
	default:
		log_error(chip, "Unknown destination register and increment setting!\n");
		return -1;
    }
	
    if(chip->mode == ATEMU_MODE_DECODE){
        if(increment){
            log_opcode(chip,  "ELPM\tR%d, Z+\n", dreg );
        }else{
            log_opcode(chip,  "ELPM\tR%d, Z\n", dreg );
        }
        return 1;
    }
	
    regz_l=chip->register_file[ATEMU_REG_Z_LOW];
    regz_h=chip->register_file[ATEMU_REG_Z_HIGH];
    rampz=*(chip->rampz);
    addr=PACK_INT_24(rampz, regz_h, regz_l);
	
    /* addr is a byte address, flash is word addressable */
    data=(chip->flash[addr >> 1]);
    if(addr & 1){
        chip->register_file[dreg]=TOP_BYTE(data);
    }else{
        chip->register_file[dreg]=BOTTOM_BYTE(data);
    }
	
    if(increment){
        /* increment addr, unpack, write back */
        addr++;
        regz_l=LOW_BYTE_24(addr);
        regz_h=MID_BYTE_24(addr);
        rampz=HIGH_BYTE_24(addr);
        chip->register_file[ATEMU_REG_Z_LOW]=regz_l;
        chip->register_file[ATEMU_REG_Z_HIGH]=regz_h;
        *(chip->rampz)=rampz;
    }
	
    chip->cycles += 3;
    chip->pc++;
	
    return 0;
}

error_t atmega_op_lds( atmega_chip *chip)
{
	error_t ret;
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "LDS\tR%d, 0x%04x\n", chip->reg_index_a, chip->addr);
        return 2;
    }
	/* if(logfile && logmemory){
	fprintf(logfile, "%03d LDS %15ld\t%x\t%d\t%x\n", chip->id, (long)chip->cycles,
	(int)chip->pc, (int)chip->reg_index_a, (unsigned int)chip->addr);
}*///whj
	
    chip->cycles+=2;
    chip->pc+=2;
	
    ret=atmega_read_sram(chip, chip->addr,
        &chip->register_file[chip->reg_index_a]);
    
	return ret;
}

error_t atmega_op_ld_elpm_pop( atmega_chip *chip )
{
    error_t result;
    /* unique on bottom nibble */
    switch(chip->bl_nibble){
	case 0x0:
		/* LDS */
		result = atmega_op_lds(chip);
		break;
	case 0x1:
	case 0x2:
	case 0x9:
	case 0xa:
	case 0xc:
	case 0xd:
	case 0xe:
		/* LD */
		result = atmega_op_ld_ldd(chip);
		break;
	case 0x4:
	case 0x5:
		/* LPM */
		result = atmega_op_lpm(chip);
		break;
	case 0x6:
	case 0x7:
		/* ELPM */
		result = atmega_op_elpm(chip);
		break;
	case 0xf:
		/* POP */
		result = atmega_op_pop(chip);
		break;
	default:
		/* bad instruction.  What do I do? */
		log_error(chip, "unknown instruction\n");
		result = -1;
		break;
    }
	
    return result;
}

error_t atmega_op_st_std(atmega_chip *chip){
    int inc;  /* amount to increment by */
    uint16_t off; /* amount to offset by */
    uint8_t reg;
    uint8_t preg_l, preg_h; /* pointer register */
    uint16_t addr;
    char preg_name;
	
    reg=chip->reg_index_a;
	
    if (((chip->bl_nibble >> 2) == 3) && (chip->th_nibble == 0x9))
    {
        /* X */
        preg_l=ATEMU_REG_X_LOW;
        preg_h=ATEMU_REG_X_HIGH;
        preg_name='X';
    }else if((chip->bl_nibble >> 3) == 1){
        /* Y */
        preg_l=ATEMU_REG_Y_LOW;
        preg_h=ATEMU_REG_Y_HIGH;
        preg_name='Y';
    }else if((chip->bl_nibble >> 3) == 0){
        /* Z */
        preg_l=ATEMU_REG_Z_LOW;
        preg_h=ATEMU_REG_Z_HIGH;
        preg_name='Z';
    }else{
        /* unknown pointer register */
        log_error(chip, "error: unknown pointer register\n");
        return -1;
    }
    
    addr=PACK_INT_16(chip->register_file[preg_h], chip->register_file[preg_l]);
	
    if(chip->th_nibble==0x8 || chip->th_nibble==0xa){
        inc=0;
        off=chip->b;
        if(chip->mode == ATEMU_MODE_DECODE){
            if(off != 0){
                log_opcode(chip, "STD\t%c%+d, R%d\n", preg_name, off, reg);
            }else{
                log_opcode(chip, "ST\t%c, R%d\n", preg_name, reg);
            }
            return 1;
        }
    }else if((chip->bl_nibble & 3) == 0){
        inc=0;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
            log_opcode(chip, "ST\t%c, R%d\n", preg_name, reg);
            return 1;
        }
    }else if((chip->bl_nibble & 3) == 1){
        inc=1;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
            log_opcode(chip, "ST\t%c+, R%d\n", preg_name, reg);
            return 1;
        }
    }else if((chip->bl_nibble & 3) == 2){
        inc=-1;
        off=0;
        if(chip->mode == ATEMU_MODE_DECODE){
            log_opcode(chip, "ST\t-%c, R%d\n", preg_name, reg);
            return 1;
        }
    }else{
        log_error(chip, "Unknown increment setting\n");
        return -1;
    }
	
    if(inc == -1){
        addr--;
    }
	
    atmega_write_sram(chip, addr + off, chip->register_file[reg]);
	
    if(inc == 1){
        addr++;
    }
	/*
    if(logfile && logmemory){
	fprintf(logfile, "%03d STD %15ld\t%x\t%d\t%x\t%x\n", chip->id, (long)chip->cycles,
	(int)chip->pc, (int)reg, (unsigned int)addr, (unsigned int)chip->register_file[reg]);
}*///whj
	
    chip->register_file[preg_h]=HIGH_BYTE_16(addr);
    chip->register_file[preg_l]=LOW_BYTE_16(addr);
	
    chip->cycles += 2;
    chip->pc++;
	
    return 0;
}

error_t atmega_op_sts(atmega_chip *chip)
{
    error_t ret;
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "STS\t0x%04x, R%d\n", chip->addr, chip->reg_index_a);
        return 2;
    }
	/* if(logfile && logmemory){
	fprintf(logfile, "%03d STS %15ld\t%x\t%d\t%x\t%x\n", chip->id, (long)chip->cycles,
	(int)chip->pc, (int)chip->reg_index_a, (unsigned int)chip->addr, (unsigned int)chip->a);
}*///whj
	
    chip->pc+=2;
    chip->cycles+=2;
    ret = atmega_write_sram(chip, chip->addr, chip->a);
	
    return ret;
}

error_t atmega_op_st_push( atmega_chip *chip )
{
    error_t result;
	
    /* unique on bottom nibble */
	
    switch(chip->bl_nibble){
	case 0x0:
		/* STS */
		result = atmega_op_sts(chip);
		break;
	case 0x1:
	case 0x2:
	case 0x9:
	case 0xa:
	case 0xc:
	case 0xd:
	case 0xe:
		/* ST X*/
		result = atmega_op_st_std(chip);
		break;
	case 0xf:
		/* PUSH */
		result = atmega_op_push(chip);
		break;
	default:
		/* not an instruction */
		log_error(chip,  "unknown instruction\n" );
		result = 0;
		break;
    }
	
    return result;
}

error_t atmega_op_mov( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "MOV\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
    
    chip->register_file[chip->reg_index_a] = chip->b;
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_movw( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "MOVW\tR%d, R%d\n", chip->reg_index_a, chip->reg_index_b);
        return 1;
    }
	
    chip->register_file[chip->reg_index_a + 1] = 
        chip->register_file[chip->reg_index_b + 1];
    chip->register_file[chip->reg_index_a] = chip->b;
    
    chip->cycles += 1;
    chip->pc++;
    return 0;
}

error_t atmega_op_spm( atmega_chip *chip )
{
    log_error(chip,  "SPM not implemented\n" );
	
    return 0;
}

error_t atmega_op_push( atmega_chip *chip )
{
    error_t result;
    uint16_t stack;
	
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "PUSH\tR%d\n", chip->reg_index_a );
        return 1;
    }
	
    stack = PACK_INT_16( *chip->sph, *chip->spl );
	
    result = atmega_write_sram( chip, stack, chip->a );
	
    if ( result < 0 )
    {
        log_error(chip,  "write_sram error.\n" );
        return result;
    }
	
    if ( stack == 0x0000 )
    {
        log_error(chip,  "stack error.\n" );
        return -EFAULT;
    }
	
    stack--;
	
    *chip->sph = HIGH_BYTE_16(stack);
    *chip->spl = LOW_BYTE_16(stack);
	
    chip->cycles += 2;
    chip->pc++;
	
    return 0;    
}

error_t atmega_op_pop( atmega_chip *chip )
{
    error_t result;
    uint16_t stack;
	
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "POP\tR%d\n", chip->reg_index_a );
        return 1;
    }
	
    stack = PACK_INT_16( *chip->sph, *chip->spl );
    stack++;
	
    result = atmega_read_sram( chip, stack, 
        &chip->register_file[chip->reg_index_a] );
	
    if ( result < 0 )
    {
        return result;
    }
	
    if ( stack == chip->pc_mask )
    {
        return -EFAULT;
    }
	
    *chip->sph = HIGH_BYTE_16(stack);
    *chip->spl = LOW_BYTE_16(stack);
	
    chip->cycles += 2;
    chip->pc++;
	
    return 0;    
}

