/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met: 
* 
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_OPCODES_H
#define ATEMU_OPCODES_H

#include "../EM/atmega128.h"

error_t atmega_push_help( atmega_chip *chip, uint32_t value );

error_t atmega_pop_help( atmega_chip *chip, uint32_t *value );

void atmega_alu_set_add_flags( atmega_chip *chip, uint8_t R, uint8_t Rd, uint8_t Rr);

void atmega_alu_set_sub_flags( atmega_chip *chip, uint8_t R, uint8_t Rd, uint8_t Rr);

void atmega_alu_set_logic_flags( atmega_chip *chip, uint8_t R);

void atmega_alu_set_sub_flagsw( atmega_chip *chip, uint16_t R, uint16_t Rd);

void atmega_alu_set_add_flagsw( atmega_chip *chip, uint16_t R, uint16_t Rd);

/*************************************************/
/* Opcode implementation functions */
/*************************************************/

/* alu */
error_t atmega_op_muls( atmega_chip *chip );
error_t atmega_op_mul( atmega_chip *chip );
error_t atmega_op_mul_group( atmega_chip *chip );
error_t atmega_op_mulsu( atmega_chip *chip );
error_t atmega_op_fmul( atmega_chip *chip );
error_t atmega_op_fmulsu( atmega_chip *chip );
error_t atmega_op_fmuls( atmega_chip *chip );
error_t atmega_op_sbc( atmega_chip *chip );
error_t atmega_op_add( atmega_chip *chip );
error_t atmega_op_sub( atmega_chip *chip );
error_t atmega_op_adc( atmega_chip *chip );
error_t atmega_op_and( atmega_chip *chip );
error_t atmega_op_eor( atmega_chip *chip );
error_t atmega_op_or( atmega_chip *chip );
error_t atmega_op_sbci( atmega_chip *chip );
error_t atmega_op_subi( atmega_chip *chip );
error_t atmega_op_ori( atmega_chip *chip );
error_t atmega_op_andi( atmega_chip *chip );
error_t atmega_op_adiw( atmega_chip *chip );
error_t atmega_op_sbiw( atmega_chip *chip );
error_t atmega_op_com( atmega_chip *chip );
error_t atmega_op_neg( atmega_chip *chip );
error_t atmega_op_inc( atmega_chip *chip );
error_t atmega_op_dec( atmega_chip *chip );

/* bit */
error_t atmega_op_bset( atmega_chip *chip );
error_t atmega_op_bclr( atmega_chip *chip );
error_t atmega_op_bld( atmega_chip *chip );
error_t atmega_op_bst( atmega_chip *chip );
error_t atmega_op_cbi( atmega_chip *chip );
error_t atmega_op_sbi( atmega_chip *chip );
error_t atmega_op_swap( atmega_chip *chip );
error_t atmega_op_asr( atmega_chip *chip );
error_t atmega_op_lsr( atmega_chip *chip );
error_t atmega_op_ror( atmega_chip *chip );

/* branch */
error_t atmega_op_flags_jump_arith( atmega_chip *chip );
error_t atmega_op_brbs( atmega_chip *chip );
error_t atmega_op_brbc( atmega_chip *chip );
error_t atmega_op_sbrc( atmega_chip *chip );
error_t atmega_op_sbrs( atmega_chip *chip );
error_t atmega_op_rjmp( atmega_chip *chip );
error_t atmega_op_rcall( atmega_chip *chip );
error_t atmega_op_sbic( atmega_chip *chip );
error_t atmega_op_sbis( atmega_chip *chip );
error_t atmega_op_cpc( atmega_chip *chip );
error_t atmega_op_cpse( atmega_chip *chip );
error_t atmega_op_cp( atmega_chip *chip );
error_t atmega_op_cpi( atmega_chip *chip );
error_t atmega_op_ret( atmega_chip *chip );
error_t atmega_op_reti( atmega_chip *chip );
error_t atmega_op_ijmp( atmega_chip *chip );
error_t atmega_op_eijmp( atmega_chip *chip );
error_t atmega_op_icall( atmega_chip *chip );
error_t atmega_op_eicall( atmega_chip *chip );

/* data */
error_t atmega_op_movw( atmega_chip *chip );
error_t atmega_op_mov( atmega_chip *chip );
error_t atmega_op_st_std( atmega_chip *chip );
error_t atmega_op_ld_ldd( atmega_chip *chip );
error_t atmega_op_lds( atmega_chip *chip );
error_t atmega_op_ld_elpm_pop( atmega_chip *chip );
error_t atmega_op_elpm( atmega_chip *chip );
error_t atmega_op_lpm( atmega_chip *chip );
error_t atmega_op_in( atmega_chip *chip );
error_t atmega_op_out( atmega_chip *chip );
error_t atmega_op_ldi( atmega_chip *chip );
error_t atmega_op_st_push( atmega_chip *chip );
error_t atmega_op_spm( atmega_chip *chip );
error_t atmega_op_pop( atmega_chip *chip );
error_t atmega_op_push( atmega_chip *chip );

/* mcu */
error_t atmega_op_nop( atmega_chip *chip );
error_t atmega_op_sleep( atmega_chip *chip );
error_t atmega_op_wdr( atmega_chip *chip );
error_t atmega_op_break( atmega_chip *chip );

#endif

